import { read } from 'jimp';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

const CHAR_SET = {
    nnnnnww: '0',
    nnnnwwn: '1',
    nnnwnnw: '2',
    wwnnnnn: '3',
    nnwnnwn: '4',
    wnnnnwn: '5',
    nwnnnnw: '6',
    nwnnwnn: '7',
    nwwnnnn: '8',
    wnnwnnn: '9',
    nnnwwnn: '-',
    nnwwnnn: '$',
    wnnnwnw: ':',
    wnwnnnw: '/',
    wnwnwnn: '.',
    nnwwwww: '+',
    nnwwnwn: 'A',
    nnnwnww: 'B',
    nwnwnnw: 'C',
    nnnwwwn: 'D',
};
function decoder(lines) {
    const code = [];
    const barThreshold = Math.ceil(lines.reduce((pre, item) => (pre + item) / 2, 0));
    // Read one encoded character at a time.
    while (lines.length > 0) {
        const seg = lines.splice(0, 8).splice(0, 7);
        const a = seg.map(line => (line < barThreshold ? 'n' : 'w')).join('');
        code.push(CHAR_SET[a]);
    }
    return code.join('');
}

const WIDTH_TBL = [
    '212222',
    '222122',
    '222221',
    '121223',
    '121322',
    '131222',
    '122213',
    '122312',
    '132212',
    '221213',
    '221312',
    '231212',
    '112232',
    '122132',
    '122231',
    '113222',
    '123122',
    '123221',
    '223211',
    '221132',
    '221231',
    '213212',
    '223112',
    '312131',
    '311222',
    '321122',
    '321221',
    '312212',
    '322112',
    '322211',
    '212123',
    '212321',
    '232121',
    '111323',
    '131123',
    '131321',
    '112313',
    '132113',
    '132311',
    '211313',
    '231113',
    '231311',
    '112133',
    '112331',
    '132131',
    '113123',
    '113321',
    '133121',
    '313121',
    '211331',
    '231131',
    '213113',
    '213311',
    '213131',
    '311123',
    '311321',
    '331121',
    '312113',
    '312311',
    '332111',
    '314111',
    '221411',
    '431111',
    '111224',
    '111422',
    '121124',
    '121421',
    '141122',
    '141221',
    '112214',
    '112412',
    '122114',
    '122411',
    '142112',
    '142211',
    '241211',
    '221114',
    '413111',
    '241112',
    '134111',
    '111242',
    '121142',
    '121241',
    '114212',
    '124112',
    '124211',
    '411212',
    '421112',
    '421211',
    '212141',
    '214121',
    '412121',
    '111143',
    '111341',
    '131141',
    '114113',
    '114311',
    '411113',
    '411311',
    '113141',
    '114131',
    '311141',
    '411131',
    '211412',
    '211214',
    '211232',
    '233111',
    '211133',
    '2331112',
];
const TBL_A = [
    ' ',
    '!',
    '"',
    '#',
    '$',
    '%',
    '&',
    "'",
    '(',
    ')',
    '*',
    '+',
    ',',
    '-',
    '.',
    '/',
    '0',
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    ':',
    ';',
    '<',
    '=',
    '>',
    '?',
    '@',
    'A',
    'B',
    'C',
    'D',
    'E',
    'F',
    'G',
    'H',
    'I',
    'J',
    'K',
    'L',
    'M',
    'N',
    'O',
    'P',
    'Q',
    'R',
    'S',
    'T',
    'U',
    'V',
    'W',
    'X',
    'Y',
    'Z',
    '[',
    '\\',
    ']',
    '^',
    '_',
    'NUL',
    'SOH',
    'STX',
    'ETX',
    'EOT',
    'ENQ',
    'ACK',
    'BEL',
    'BS',
    'HT',
    'LF',
    'VT',
    'FF',
    'CR',
    'SO',
    'SI',
    'DLE',
    'DC1',
    'DC2',
    'DC3',
    'DC4',
    'NAK',
    'SYN',
    'ETB',
    'CAN',
    'EM',
    'SUB',
    'ESC',
    'FS',
    'GS',
    'RS',
    'US',
    'FNC 3',
    'FNC 2',
    'Shift B',
    'Code C',
    'Code B',
    'FNC 4',
    'FNC 1',
    'Code A',
    'Code B',
    'Code C',
    'Stop',
    'Reverse Stop',
];
const TBL_B = [
    ' ',
    '!',
    '"',
    '#',
    '$',
    '%',
    '&',
    "'",
    '(',
    ')',
    '*',
    '+',
    ',',
    '-',
    '.',
    '/',
    '0',
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    ':',
    ';',
    '<',
    '=',
    '>',
    '?',
    '@',
    'A',
    'B',
    'C',
    'D',
    'E',
    'F',
    'G',
    'H',
    'I',
    'J',
    'K',
    'L',
    'M',
    'N',
    'O',
    'P',
    'Q',
    'R',
    'S',
    'T',
    'U',
    'V',
    'W',
    'X',
    'Y',
    'Z',
    '[',
    '\\',
    ']',
    '^',
    '_',
    '`',
    'a',
    'b',
    'c',
    'd',
    'e',
    'f',
    'g',
    'h',
    'i',
    'j',
    'k',
    'l',
    'm',
    'n',
    'o',
    'p',
    'q',
    'r',
    's',
    't',
    'u',
    'v',
    'w',
    'x',
    'y',
    'z',
    '{',
    '|',
    '}',
    '~',
    'DEL',
    'FNC 3',
    'FNC 2',
    'Shift A',
    'Code C',
    'FNC 4',
    'Code A',
    'FNC 1',
    'Code A',
    'Code B',
    'Code C',
    'Stop',
    'Reverse Stop',
];
const TBL_C = [
    '00',
    '01',
    '02',
    '03',
    '04',
    '05',
    '06',
    '07',
    '08',
    '09',
    '10',
    '11',
    '12',
    '13',
    '14',
    '15',
    '16',
    '17',
    '18',
    '19',
    '20',
    '21',
    '22',
    '23',
    '24',
    '25',
    '26',
    '27',
    '28',
    '29',
    '30',
    '31',
    '32',
    '33',
    '34',
    '35',
    '36',
    '37',
    '38',
    '39',
    '40',
    '41',
    '42',
    '43',
    '44',
    '45',
    '46',
    '47',
    '48',
    '49',
    '50',
    '51',
    '52',
    '53',
    '54',
    '55',
    '56',
    '57',
    '58',
    '59',
    '60',
    '61',
    '62',
    '63',
    '64',
    '65',
    '66',
    '67',
    '68',
    '69',
    '70',
    '71',
    '72',
    '73',
    '74',
    '75',
    '76',
    '77',
    '78',
    '79',
    '80',
    '81',
    '82',
    '83',
    '84',
    '85',
    '86',
    '87',
    '88',
    '89',
    '90',
    '91',
    '92',
    '93',
    '94',
    '95',
    '96',
    '97',
    '98',
    '99',
    'Code B',
    'Code A',
    'FNC 1',
    'Code A',
    'Code B',
    'Code C',
    'Stop',
    'Reverse Stop',
];
const computeGroup = (lines) => {
    const count = lines.length - 13;
    const factor = lines.reduce((pre, item, i) => {
        if (i >= count)
            return pre;
        return pre + item;
    }, 0) /
        (Math.ceil(count / 6) * 11);
    console.log(factor);
    return lines.map(item => Math.round(item / factor) || 1);
};
function decoder$1(lines) {
    const code = [];
    let lookupTBL = TBL_B;
    // let sumOP = 0
    let letterKey;
    let letterCode;
    let letterCodePrev;
    let keyIndex;
    const computedLines = computeGroup(lines);
    if (!computedLines)
        return '';
    // extract terminal bar
    computedLines.pop();
    // skip check code and stop code using -13
    for (let i = 0; i * 6 < computedLines.length - 13; i += 1) {
        letterKey = computedLines.slice(i * 6, (i + 1) * 6).join('');
        keyIndex = WIDTH_TBL.indexOf(letterKey);
        letterCode = lookupTBL[keyIndex];
        // sumOP += i * keyIndex
        switch (letterCode) {
            case 'Code A':
                lookupTBL = TBL_A;
                break;
            case 'Code B':
                lookupTBL = TBL_B;
                break;
            case 'Code C':
                lookupTBL = TBL_C;
                break;
            case 'FNC 4':
                break;
            default:
                if (letterCode) {
                    if (letterCodePrev === 'FNC 4') {
                        code.push(letterCode.charCodeAt(0) + 128);
                    }
                    else {
                        code.push(letterCode);
                    }
                    letterCodePrev = letterCode;
                }
                else {
                    code.push('?');
                }
                break;
        }
    }
    // letterKey = computedLines.slice(0, 6).join('')
    // if (sumOP % 103 !== WIDTH_TBL.indexOf(letterKey)) return ''
    console.log('CODE: ', code.join(''));
    return code.join('');
}

const CHAR_SET$1 = {
    nnnwwnwnn: '0',
    wnnwnnnnw: '1',
    nnwwnnnnw: '2',
    wnwwnnnnn: '3',
    nnnwwnnnw: '4',
    wnnwwnnnn: '5',
    nnwwwnnnn: '6',
    nnnwnnwnw: '7',
    wnnwnnwnn: '8',
    nnwwnnwnn: '9',
    wnnnnwnnw: 'A',
    nnwnnwnnw: 'B',
    wnwnnwnnn: 'C',
    nnnnwwnnw: 'D',
    wnnnwwnnn: 'E',
    nnwnwwnnn: 'F',
    nnnnnwwnw: 'G',
    wnnnnwwnn: 'H',
    nnwnnwwnn: 'I',
    nnnnwwwnn: 'J',
    wnnnnnnww: 'K',
    nnwnnnnww: 'L',
    wnwnnnnwn: 'M',
    nnnnwnnww: 'N',
    wnnnwnnwn: 'O',
    nnwnwnnwn: 'P',
    nnnnnnwww: 'Q',
    wnnnnnwwn: 'R',
    nnwnnnwwn: 'S',
    nnnnwnwwn: 'T',
    wwnnnnnnw: 'U',
    nwwnnnnnw: 'V',
    wwwnnnnnn: 'W',
    nwnnwnnnw: 'X',
    wwnnwnnnn: 'Y',
    nwwnwnnnn: 'Z',
    nwnnnnwnw: '-',
    wwnnnnwnn: '.',
    nwwnnnwnn: ' ',
    nwnwnwnnn: '$',
    nwnwnnnwn: '/',
    nwnnnwnwn: '+',
    nnnwnwnwn: '%',
    nwnnwnwnn: '*',
};
function decoder$2(lines) {
    const code = [];
    const barThreshold = Math.ceil(lines.reduce((pre, item) => pre + item, 0) / lines.length);
    // Read one encoded character at a time.
    while (lines.length > 0) {
        const sequenceBar = lines
            .splice(0, 10)
            .map(line => (line > barThreshold ? 'w' : 'n'))
            .slice(0, 9)
            .join('');
        code.push(CHAR_SET$1[sequenceBar]);
    }
    if (code.pop() !== '*' || code.shift() !== '*')
        return '';
    return code.join('');
}

const CHAR_SET$2 = [
    { '100010100': '0' },
    { '101001000': '1' },
    { '101000100': '2' },
    { '101000010': '3' },
    { '100101000': '4' },
    { '100100100': '5' },
    { '100100010': '6' },
    { '101010000': '7' },
    { '100010010': '8' },
    { '100001010': '9' },
    { '110101000': 'A' },
    { '110100100': 'B' },
    { '110100010': 'C' },
    { '110010100': 'D' },
    { '110010010': 'E' },
    { '110001010': 'F' },
    { '101101000': 'G' },
    { '101100100': 'H' },
    { '101100010': 'I' },
    { '100110100': 'J' },
    { '100011010': 'K' },
    { '101011000': 'L' },
    { '101001100': 'M' },
    { '101000110': 'N' },
    { '100101100': 'O' },
    { '100010110': 'P' },
    { '110110100': 'Q' },
    { '110110010': 'R' },
    { '110101100': 'S' },
    { '110100110': 'T' },
    { '110010110': 'U' },
    { '110011010': 'V' },
    { '101101100': 'W' },
    { '101100110': 'X' },
    { '100110110': 'Y' },
    { '100111010': 'Z' },
    { '100101110': '-' },
    { '111010100': '.' },
    { '111010010': ' ' },
    { '111001010': '$' },
    { '101101110': '/' },
    { '101110110': '+' },
    { '110101110': '%' },
    { '100100110': '($)' },
    { '111011010': '(%)' },
    { '111010110': '(/)' },
    { '100110010': '(+)' },
    { '101011110': '*' },
];
function decoder$3(lines) {
    const code = [];
    const binary = [];
    // remove termination bar
    lines.pop();
    const barThreshold = Math.ceil(lines.reduce((pre, item) => pre + item, 0) / lines.length);
    const minBarWidth = Math.ceil(lines.reduce((pre, item) => {
        if (item < barThreshold)
            return (pre + item) / 2;
        return pre;
    }, 0));
    // leave the padded *
    for (let i = 0; i < lines.length; i += 1) {
        let segment = lines[i];
        while (segment > 0) {
            if (i % 2 === 0) {
                binary.push(1);
            }
            else {
                binary.push(0);
            }
            segment -= minBarWidth;
        }
    }
    for (let i = 0; i < binary.length; i += 9) {
        const searcKey = binary.slice(i, i + 9).join('');
        const char = CHAR_SET$2.filter(item => Object.keys(item)[0] === searcKey);
        code.push(char[0][searcKey]);
    }
    if (code.shift() !== '*' || code.pop() !== '*')
        return '';
    const K = code.pop();
    let sum = 0;
    let letter;
    let Value;
    const findValue = (item) => Object.values(item)[0] === letter;
    for (let i = code.length - 1; i >= 0; i -= 1) {
        letter = code[i];
        Value = CHAR_SET$2.indexOf(CHAR_SET$2.filter(findValue)[0]);
        sum += Value * (1 + ((code.length - (i + 1)) % 20));
    }
    if (Object.values(CHAR_SET$2[sum % 47])[0] !== K)
        return '';
    const C = code.pop();
    sum = 0;
    for (let i = code.length - 1; i >= 0; i -= 1) {
        letter = code[i];
        Value = CHAR_SET$2.indexOf(CHAR_SET$2.filter(findValue)[0]);
        sum += Value * (1 + ((code.length - (i + 1)) % 20));
    }
    if (Object.values(CHAR_SET$2[sum % 47])[0] !== C)
        return '';
    return code.join('');
}

const CHAR_SET$3 = [
    'nnwwn',
    'wnnnw',
    'nwnnw',
    'wwnnn',
    'nnwnw',
    'wnwnn',
    'nwwnn',
    'nnnww',
    'wnnwn',
    'nwnwn',
];
function decoder$4(lines, type) {
    const code = [];
    const barThreshold = Math.ceil(lines.reduce((pre, item) => (pre + item) / 2, 0));
    if (type === 'interleaved') {
        // extract start/ends pair
        const startChar = lines
            .splice(0, 4)
            .map((line) => (line > barThreshold ? 'w' : 'n'))
            .join('');
        const endChar = lines
            .splice(lines.length - 3, 3)
            .map((line) => (line > barThreshold ? 'w' : 'n'))
            .join('');
        if (startChar !== 'nnnn' || endChar !== 'wnn')
            return '';
        // Read one encoded character at a time.
        while (lines.length > 0) {
            const seg = lines.splice(0, 10);
            const a = seg
                .filter((item, index) => index % 2 === 0)
                .map(line => (line > barThreshold ? 'w' : 'n'))
                .join('');
            code.push(CHAR_SET$3.indexOf(a));
            const b = seg
                .filter((item, index) => index % 2 !== 0)
                .map(line => (line > barThreshold ? 'w' : 'n'))
                .join('');
            code.push(CHAR_SET$3.indexOf(b));
        }
    }
    else {
        // extract start/ends pair
        const startChar = lines
            .splice(0, 6)
            .filter((item, index) => index % 2 === 0)
            .map((line) => (line > barThreshold ? 'w' : 'n'))
            .join('');
        const endChar = lines
            .splice(lines.length - 5, 5)
            .filter((item, index) => index % 2 === 0)
            .map((line) => (line > barThreshold ? 'w' : 'n'))
            .join('');
        if (startChar !== 'wwn' || endChar !== 'wnw')
            return '';
        // Read one encoded character at a time.
        while (lines.length > 0) {
            const a = lines
                .splice(0, 10)
                .filter((item, index) => index % 2 === 0)
                .map(line => (line > barThreshold ? 'w' : 'n'))
                .join('');
            code.push(CHAR_SET$3.indexOf(a));
        }
    }
    return code.join('');
}

const UPC_SET = {
    '3211': '0',
    '2221': '1',
    '2122': '2',
    '1411': '3',
    '1132': '4',
    '1231': '5',
    '1114': '6',
    '1312': '7',
    '1213': '8',
    '3112': '9',
};
function decoder$5(lines) {
    let code = '';
    // start indicator/reference lines
    const bar = (lines[0] + lines[1] + lines[2]) / 3;
    // remove start pattern
    lines.shift();
    lines.shift();
    lines.shift();
    // remove end pattern
    lines.pop();
    lines.pop();
    lines.pop();
    // remove middle check pattern
    lines.splice(24, 5);
    for (let i = 0; i < lines.length; i += 4) {
        const group = lines.slice(i, i + 4);
        const digits = [group[0] / bar, group[1] / bar, group[2] / bar, group[3] / bar].map(digit => digit === 1.5 ? 1 : Math.round(digit));
        const result = UPC_SET[digits.join('')] || UPC_SET[digits.reverse().join('')];
        if (result) {
            code += result;
        }
        else {
            code += '?';
        }
    }
    return code;
}

const UPC_SET$1 = {
    '3211': '0',
    '2221': '1',
    '2122': '2',
    '1411': '3',
    '1132': '4',
    '1231': '5',
    '1114': '6',
    '1312': '7',
    '1213': '8',
    '3112': '9',
};
function decoder$6(lines) {
    let code = '';
    // start indicator/reference lines
    const bar = (lines[0] + lines[1] + lines[2]) / 3;
    // remove start pattern
    lines.shift();
    lines.shift();
    lines.shift();
    // remove end pattern
    lines.pop();
    lines.pop();
    lines.pop();
    // remove middle check pattern
    lines.splice(16, 5);
    for (let i = 0; i < lines.length; i += 4) {
        const group = lines.slice(i, i + 4);
        const digits = [group[0] / bar, group[1] / bar, group[2] / bar, group[3] / bar].map(digit => digit === 1.5 ? 1 : Math.round(digit));
        const result = UPC_SET$1[digits.join('')] || UPC_SET$1[digits.reverse().join('')];
        if (result) {
            code += result;
        }
        else {
            code += '?';
        }
    }
    return code;
}

function applyAdaptiveThreshold(data, width, height) {
    const integralImage = new Array(width * height).fill(0);
    const channels = data.length / (width * height);
    const t = 0.15; // threshold percentage
    const s = Math.floor(height); // bracket size
    const s2 = Math.floor(s / 2);
    for (let i = 0; i < width; i += 1) {
        let sum = 0;
        for (let j = 0; j < height; j += 1) {
            const pureIndex = j * width + i;
            const index = pureIndex * channels;
            // greyscale
            const v = (data[index] + data[index + 1] + data[index + 2]) / 3;
            data[index] = v;
            data[index + 1] = v;
            data[index + 2] = v;
            sum += v;
            if (i === 0) {
                integralImage[pureIndex] = sum;
            }
            else {
                integralImage[pureIndex] = integralImage[pureIndex - 1] + sum;
            }
        }
    }
    // skip edge rows
    for (let i = 0; i < width; i += 1) {
        for (let j = 0; j < height; j += 1) {
            const pureIndex = j * width + i;
            const index = pureIndex * channels;
            // no. of pixels per window
            let x1 = i - s2;
            let x2 = i + s2;
            let y1 = j - s2;
            let y2 = j + s2;
            if (x1 < 0)
                x1 = 0;
            if (x2 >= width)
                x2 = width - 1;
            if (y1 < 0)
                y1 = 0;
            if (y2 >= height)
                y2 = height - 1;
            const count = (x2 - x1) * (y2 - y1);
            const sum = integralImage[y2 * width + x2] -
                integralImage[y1 * width + x2] -
                integralImage[y2 * width + x1] +
                integralImage[y1 * width + x1];
            let v = 255;
            if (data[index] * count < sum * (1 - t)) {
                v = 0;
            }
            data[index] = v;
            data[index + 1] = v;
            data[index + 2] = v;
        }
    }
    return data;
}

function combineAllPossible(finalResult, result) {
    if (finalResult === '' || result === '') {
        return result;
    }
    const finalResultArr = finalResult.split('');
    const resultArr = result.split('');
    resultArr.forEach((char, index) => {
        if (!finalResultArr[index] || finalResultArr[index] === '?') {
            if (char && char !== '?') {
                finalResultArr[index] = char;
            }
        }
    });
    return finalResultArr.join('');
}

function createImageData(image) {
    const canvas = document.createElement('canvas');
    const ctx = canvas.getContext('2d');
    if (!ctx)
        throw new Error('canvas not available');
    const width = image.naturalWidth;
    const height = image.naturalHeight;
    canvas.width = width;
    canvas.height = height;
    ctx.drawImage(image, 0, 0);
    return ctx.getImageData(0, 0, width, height);
}

function isUrl(s) {
    const regexp = /(ftp|http|https):\/\/(\w+:{0,1}\w*@)?(\S+)(:[0-9]+)?(\/|\/([\w#!:.?+=&%@!\-/]))?/;
    if (s.startsWith('#'))
        return false;
    return regexp.test(s);
}

const isNode = typeof process === 'object' && process.release && process.release.name === 'node';
function getImageDataFromSource(source) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            if (typeof source === 'string') {
                if (source.startsWith('#')) {
                    const imageElement = document.getElementById(source.substr(1));
                    if (imageElement instanceof HTMLImageElement) {
                        resolve(createImageData(imageElement));
                    }
                    if (imageElement instanceof HTMLCanvasElement) {
                        const ctx = imageElement.getContext('2d');
                        if (!ctx)
                            throw new Error('Cannot create canvas 2d context');
                        resolve(ctx.getImageData(0, 0, imageElement.width, imageElement.height));
                    }
                    reject(new Error('Invalid image source specified!'));
                }
                else if (isUrl(source)) {
                    const img = new Image();
                    img.onerror = reject;
                    img.onload = () => resolve(createImageData(img));
                    img.src = source;
                }
                else if (isNode) {
                    read(source, (err, image) => {
                        if (err) {
                            reject(err);
                        }
                        else {
                            const { data, width, height } = image.bitmap;
                            resolve({
                                data: Uint8ClampedArray.from(data),
                                width,
                                height,
                            });
                        }
                    });
                }
            }
            else if (source instanceof HTMLImageElement) {
                resolve(createImageData(source));
            }
            else if (source instanceof HTMLCanvasElement) {
                const ctx = source.getContext('2d');
                if (!ctx)
                    throw new Error('Cannot create canvas 2d context');
                resolve(ctx.getImageData(0, 0, source.width, source.height));
            }
        });
    });
}

function getBitData(data, width, height) {
    const bitData = [];
    const channels = data.length / (width * height);
    for (let i = 0; i < data.length; i += channels) {
        let average = Math.sqrt((Math.pow(data[i], 2) + Math.pow(data[i + 1], 2) + Math.pow(data[i + 2], 2)) / 3);
        if (channels === 4) {
            average *= data[3] / 255;
        }
        bitData.push(average);
    }
    return bitData;
}

function applyMedianFilter(data, width, height, windowSize = 3) {
    const limit = (windowSize - 1) / 2;
    const filterWindow = [];
    for (let i = limit * -1; i < limit + 1; i += 1) {
        for (let j = limit * -1; j < limit + 1; j += 1) {
            filterWindow.push([i, j]);
        }
    }
    for (let col = limit; col < width - limit; col += 1) {
        for (let row = limit; row < height - limit; row += 1) {
            const index = row * width + col;
            const collection = [];
            for (let z = 0; z < filterWindow.length; z += 1) {
                const i = (row + filterWindow[z][0]) * width + (col + filterWindow[z][1]);
                collection.push(data[i]);
            }
            const sorted = collection.sort((a, b) => b - a);
            data[index] = sorted[Math.floor(sorted.length / 2)];
        }
    }
    return data;
}

function getLines(data, width, height) {
    let bitData = getBitData(data, width, height);
    const lines = [];
    let count = 0;
    let columnAverageLast = 0;
    bitData = applyMedianFilter(bitData, width, height);
    for (let column = 0; column < width; column += 1) {
        let columnSum = 0;
        let columnAverage = 0;
        for (let row = 0; row < height; row += 1) {
            const index = row * width + column;
            columnSum += bitData[index];
        }
        // pixels are same in column
        columnAverage = columnSum / height >= 127 ? 255 : 0;
        // skip white padding in the start & end
        if (columnAverage === 255 && count === 0)
            continue;
        // count line width
        if (columnAverage === columnAverageLast) {
            count += 1;
        }
        else {
            lines.push(count);
            columnAverageLast = columnAverage;
            count = 1;
        }
        // skip padding in the last
        if (column === width - 1 && columnAverage === 0) {
            lines.push(count);
        }
    }
    return lines;
}

const isTestEnv = process && process.env.NODE_ENV === 'test';
// TODO: add fuse.js for fuzzy search for bad quality barcodes
var BARCODE_DECODERS;
(function (BARCODE_DECODERS) {
    BARCODE_DECODERS["code-128"] = "code-128";
    BARCODE_DECODERS["code-2of5"] = "code-2of5";
    BARCODE_DECODERS["code-39"] = "code-39";
    BARCODE_DECODERS["code-93"] = "code-93";
    BARCODE_DECODERS["ean-13"] = "ean-13";
    BARCODE_DECODERS["ean-8"] = "ean-8";
    BARCODE_DECODERS["codabar"] = "codabar";
})(BARCODE_DECODERS || (BARCODE_DECODERS = {}));
function isImageLike(object) {
    return object.data && object.width && object.height;
}
function javascriptBarcodeReader({ image, barcode, barcodeType, options, }) {
    return __awaiter(this, void 0, void 0, function* () {
        let decoder$7;
        switch (barcode) {
            case BARCODE_DECODERS.codabar:
                decoder$7 = decoder;
                break;
            case BARCODE_DECODERS['code-128']:
                decoder$7 = decoder$1;
                break;
            case BARCODE_DECODERS['code-39']:
                decoder$7 = decoder$2;
                break;
            case BARCODE_DECODERS['code-93']:
                decoder$7 = decoder$3;
                break;
            case BARCODE_DECODERS['code-2of5']:
                decoder$7 = decoder$4;
                break;
            case BARCODE_DECODERS['ean-13']:
                decoder$7 = decoder$5;
                break;
            case BARCODE_DECODERS['ean-8']:
                decoder$7 = decoder$6;
                break;
            default:
                throw new Error(`Invalid barcode specified. Available decoders: ${BARCODE_DECODERS}.`);
        }
        const useSinglePass = isTestEnv || (options && options.singlePass) || false;
        const imageData = isImageLike(image) ? image : yield getImageDataFromSource(image);
        const width = imageData.width;
        const height = imageData.height;
        const channels = imageData.data.length / (width * height);
        let finalResult = '';
        // apply adaptive threshold
        if (options && options.useAdaptiveThreshold) {
            applyAdaptiveThreshold(imageData.data, width, height);
        }
        // check points for barcode location
        const sPoints = [5, 6, 4, 7, 3, 8, 2, 9, 1];
        const slineStep = Math.round(height / sPoints.length);
        const rowsToScan = Math.min(3, height);
        for (let i = 0; i < sPoints.length; i += 1) {
            const sPoint = sPoints[i];
            const start = channels * width * Math.floor(slineStep * sPoint);
            const end = start + rowsToScan * channels * width;
            const lines = getLines(imageData.data.slice(start, end), width, rowsToScan);
            if (lines.length === 0) {
                if (useSinglePass)
                    throw new Error('Failed to extract barcode!');
                continue;
            }
            // Run the decoder
            const result = decoder$7(lines, barcodeType);
            if (useSinglePass)
                return result;
            if (!result)
                continue;
            if (!result.includes('?'))
                return result;
            finalResult = combineAllPossible(finalResult, result);
            if (!finalResult.includes('?'))
                return finalResult;
            if (i === sPoints.length - 1)
                return finalResult;
        }
        throw new Error('Failed to extract barcode!');
    });
}

export { BARCODE_DECODERS, javascriptBarcodeReader };
//# sourceMappingURL=javascript-barcode-reader.es5.js.map
